<?php if(!defined('PmWiki')) exit();

/*

    Copyright 2007 Michael P Majorowicz
    This file is router.php; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.  
 
    file router.php
    Brief router allows a website's url structure to be different from PmWiki's group/page structure.
  
    See: http://www.pmwiki.org/wiki/Cookbook/Router for installation and usage instructions
    
      
    CHANGELOG:
    
    by CarlosAB - 2017.05.16
    
    - Changed the script so RouterPreProcess(), now RouterStart()
      has to be called from config, so you can turn it on and off
      when needed, comment, control it better and so on.
    - An array has to be made as usual and will activate router 
      when called early or before the include with $Routes
      array, as a function option to RouterStart() + pagename.
    - Changed $aRouter to $Router, usually it activates the recipe 
      using an array called $Routes.
    - Changed $aline to $line, for readability as line is used always
      inside functions.
    - Changed "RT_" key for namespace to "RouterFn", to
      ease comprehension for new comers and still use the same namespace.
    - Changed function names and calling to use Router namespace.
    - Included RouterRedirect function.
    - Changed rline to line for uniformity on Router Redirect and ease
      function comprehension.
    - Included var RouterDashedToCamelReserved and var RouterCamelToDashedReserved
      as an array that when filled with replace pattern as key and replacement as value.
    - we splited the original CameCase function we separated ones CamelToDashedLowerCase
      and DashedLowerToCamelCase.
    - If n is empty pagename will not get filled so we resolve this with MakePageName
      and setting vars that depend on pagename and pagename itself inside RouterStart.
    - Included FmtPV['RouterOriginal'] and FmtPV['RouterProcessed'] to access
      Requested var pagename and pagename after routing process.
    - On RouterFnCamelToDashedLowerCase we use a special new notation for unicode
      to find CamelCased words "Include Space" between them.
    
    It works very well in my tests, if you find any problems post
    a comment to the talk page.
    

    DOCUMENTATION (with changes by CarlosAB)

    Questions answered by this recipe:
    
    - How can I make PmWiki urls look different?
    - How can I make PmWiki fit into a pre-existing url structure?
    - How can I make PmWiki use dashed-lower-case urls instead of CamelCaseUrls?


    Examples:
    
      Short Format:
        $Routes['MatchWiki']='MatchUrl';
       

      Long Format:
        $Routes[]=array(
           'Mode'=>'BeginsWith',
           'PageName'=>'MatchWiki',
           'Url'=>'MatchUrl'
       );
    
    
    Modes that currently exist are:

    - Symetrical Modes
    -- BeginsWith
    -- Contains
    -- EndsWith
    -- ExactMatch
    - Asymetrical Modes
    -- RegReplace
    - Special Modes
    -- CamelCaseToLowerCase


    Asymetrical Modes:

    Asymetrical modes only work in one direction. The direction
    to apply the routing instruction to is specified by adding
    another element.
    
      'Direction'=>'Out' or
      'Direction'=>'In'
      
    The 'Direction' element can be used with any instruction mode.
    
      RegReplace Example

        $Routes[]=array(
          'Mode'=>'RegReplace',
          'Search'=>'/^MatchUrl/',
          'Replace'=>'MatchWiki',
          'Direction'=>'In'
       );
       
       For inbound links,   
    
      http://example.com/MatchUrlGroup/MatchUrlPage 
      
                translates to

      MatchWikiGroup/MatchUrlPage
      
      
    But this instruction is limited to only inbound links by the ,
    'Direction'=>'In' part of the routing instruction. It is skipped
    when translating pagenames to urls for outbound links because the
    regex doesn't translate the urls correctly in the reverse direction.
    
    So a related outbound instruction is needed.
    
    $Routes[]=array(
      'Mode'=>'RegReplace',
      'Search'=>'/^MatchPage/',
      'Replace'=>'MatchUrl',
      'Direction'=>'Out'
    );
    
    Now MatchPageGroup/MatchPagePage correctly translates to
   
    http://example.com/MatchUrlGroup/MatchPagePage
   
   
    The RegReplace instruction simply passes the values of 'Search' and
    'Replace' to the php preg_replace() function. If you specify a 'Limit'
    element in the routing instruction, that value will also be passed
    to preg_replace().
    
    
    Special Modes
    
    Special modes mess with urls in an extra special way. :)
    
      CamelCaseToLowerCase Example

        The 'CamelCaseToLowerCase' mode changes the entire pmwiki url
        structure from CamelCaseUrls to dashed-lower-case-urls. It 
        functions in both directions unless you specifically add a
        'Direction' element to the routing instruction.
        

           
        http://example.com/url-group/url-page 
        
        translates to

        UrlGroup/UrlPage
        
        and vice versa
        
        When using CamelToDashedLowerCase in conjunction with the Apache
        mod-rewrite module, be sure to change your .htaccess to send all
        urls that start with lower case characters to index.php.

    Create your own routing modes
    
    Add a function named RouterFnModeName to config.php before including router.php.
    The first parameter is the path and the second parameter is an array that contains
    a single routing instruction.
    
    The 'Direction' element is always present and indicates whether the path you are
    working on is for an inbound or outbound link. Router will add the 'Direction' element
    to the instruction array if it doesn't already exist. Be sure to return $path even if
    you don't change its value.
    
    Values passed to the function in $line['Direction'] are either 'In' or 'Out'.
    The rest of the $aline array keys are determined by your Router function.
    
    The following function creates a routing mode named "Redirect"
    
      $Routes[]=array(
        'Mode'=>'Redirect',
        'In'=>'Main/HomePage',
        'Out'=>'http://example.com/'
      );


    ## Http Redirect to specified location when path matches
    # $path: page name at various stages of translation in either Wiki or Url format.
    # $rline: array containing a single routing instruction.

    function RouterRedirect($path,$line) {
      if('In'!=$line['Direction']){
        return $path;
       }
      if($path == @$line['In']){
        header('Location: '.@$line['Out'],TRUE,301);
      }
      return $path;
    }
    
    
*/


$RecipeInfo['Router']['Version']  =  '20170517';

$FmtPV['$PageUrl'] = 
  'RoutedUrl($group,$name,"FmtPv[PageUrl]")';
$FmtPV['$GroupUrl'] = 
  'RoutedGroupUrl($group,"FmtPv[GroupUrl]")';

SDV($RouterDashedToCamelReserved,array(
  'pmwiki' => 'PmWiki',
  'Pmwiki' => 'PmWiki',
  'changelog' => 'ChangeLog',
));

SDV($RouterCamelToDashedReserved,array(
  'f-a-q' => 'FAQ',
  'c-ds' => 'CDs',
  'pm-wiki' => 'pmwiki',
  'Pmwiki' => 'pmwiki',
  'change-log' => 'changelog',
));

/// Controls processing of in/out-bound links
function RouterStart($r,$p) { # r=routes & p=pagename
  global $RouterPageName,$RouterOriginal,$Cursor,
    $FmtPV,$PostConfig,$Router,$pagename;
  if($p == '') 
    $p = MakePageName("$p.$p");
  if($pagename == '')
    $pagename = $p;
  $Router = $r;
  $RouterOriginal = $p;
  $RouterPageName = RouterLoop($p,'In');
  # start changing previous pagename settings
  $_GET['n'] = $RouterPageName;
  $_GET['pagename'] = $RouterPageName;
  $_REQUEST['n'] = $RouterPageName;
  $_REQUEST['pagename'] = $RouterPageName;
  $FmtPV['$RequestedPage'] = "'".PHSC($RouterPageName, ENT_QUOTES)."'";
  $Cursor['*'] = $RouterPageName;
  # finish changing previous pagename settings
  $FmtPV['$RouterOriginal'] = 
          "'".$RouterOriginal."'";
  $FmtPV['$RouterProcessed'] = 
          "'".$RouterPageName."'";
  $Router = array_reverse($r,TRUE);
}

/// Converts a group name into an outbound 
/// link for $GroupUrl page variable
function RoutedGroupUrl($group,$call='') {
  global $DefaultGroup,$DefaultName;
  if($group == $DefaultGroup) {
    return RoutedUrl($DefaultGroup,$DefaultName,$call);
  }
  return RoutedUrl($group,$group);
}

/// Converts a group/page pair into an outbound link for 
/// $PageUrl page variable caches routed urls in a static array
function RoutedUrl($group,$name,$call='') {
  static $RouterUrlCache = array();
  global $Router,$EnablePathInfo,$ScriptUrl;
  $path = "$group.$name"; 

  if($EnablePathInfo) {
    $path = "$group/$name";
  }
  $pathkey = $path; #pathkey also used for saving to cache below
  if(array_key_exists($pathkey,$RouterUrlCache)) {
    return $RouterUrlCache[$pathkey];
  }
  $path = RouterLoop($path,'Out');
  $url = $ScriptUrl.'?n='.$path;
  if($EnablePathInfo) {
    $url = $ScriptUrl.'/'.$path;
  }
  if('' == $path) {
    $url = $ScriptUrl;
  }
  $RouterUrlCache[$pathkey] = PUE($url);
  return $RouterUrlCache[$pathkey];
}

/// Loops through configured routing instructions 
/// for both inbound and outbound links
function RouterLoop($path,$direction) {
  global $Router,$RouterDebug;
  
  foreach ($Router as $k => $v) {
    $line = array();
    if(is_array($v)){
      $line = $v;
      #is instuction limiting processing to one direction
      if (array_key_exists('Direction',$line)){
        if ($direction != $line['Direction']){
          continue;
        }
      }
    }else{
      #default to basic instruction format
      $line['Mode'] = 'BeginsWith';
      $line['PageName'] = $k;
      $line['Url'] = $v;
    }
    $line['Direction'] = $direction;
    if(!array_key_exists('Mode',$line)){
      continue;
    }
    $modefunc = 'RouterFn'.$line['Mode'];
    if(!function_exists($modefunc)){
      continue;
    }
    $path = $modefunc($path,$line);
  }
  return $path;
}

/// Determines search and replace values based on 
/// inbound/outbound direction of the link being processed
function RouterFnGetSearchReplace($line){
  if('Out' == @$line['Direction']){
    return array(
      @$line['PageName'], #search
      @$line['Url'] #replace
    );
  }
  return array(
    @$line['Url'], #search
    @$line['PageName'] #replace
  );
}

/// For a match, the search value must start
/// at the beginning of the path.
function RouterFnBeginsWith($path,$line){
  list($search,$replace) = RouterGetSearchReplace($line);
  if(strpos($path,$search) === 0){
    $path = substr_replace($path,
       $replace,0,strlen($search));
  }
  return $path;
}

/// For a match, the search value can be
/// found anywhere in the path
function RouterFnContains($path,$line){
  list($search,$replace) = RouterGetSearchReplace($line);
  $path = str_replace($search,$replace,$path);
  return $path;
}

/// For a match, the search value must be at the end of the path
function RouterFnEndsWith($path,$line){
  list($search,$replace) = RouterGetSearchReplace($line);
  if(substr($path,-strlen($search)) === $search) {
    $path = substr_replace($path,
       $replace,-strlen($search),strlen($search));
  }
  return $path;
}

/// For a match, the search value must
/// be exactly equal to the path
function RouterFnExactMatch($path,$line){
  list($search,$replace) = RouterGetSearchReplace($line);
  if($path == $search) {
    return $replace;
  }
  return $path;
}

/// Process a search/replace pair with preg_replace()
function RouterFnRegReplace($path,$line){
  if(!array_key_exists('Search',$line) ||
       !array_key_exists('Replace',$line)) {
    return $path;
  }
  $limit = -1;
  if(array_key_exists('Limit',$line)) {
    $limit = $line['Limit'];
  }
  $path = preg_replace($line['Search'],
     $line['Replace'],$path,$limit);
  return $path;
}

function RouterFnDashedLowerToCamelCase($path,$line,$call=''){
global $pagename, $EnablePathInfo,$RouterDashedToCamelReserved;
  $fstpath = $path;
  if(strpos('-',$path) !== false) return $path;
  $path = strtr($path,array(
    '-' => ' ','/' => ' / ','.' => ' . '));
  $path = ucwords($path);
  $path = strtr($path,array(
    ' . ' => '.',' / ' =>'/',' ' =>''));
  if($EnablePathInfo == '1')
    $path = str_replace('.','/',$path);
  $path = strtr($path,array('.-' => '.','/-' => '/'));
  $path = strtr($path,$RouterDashedToCamelReserved);
  return $path;
}

/// Converts between lower-case-urls and CamelCase page names
function RouterFnCamelToDashedLowerCase($path,$line,$call=''){
global $EnablePathInfo,$RouterCamelToDashedReserved;
  $fstpath = $path;
  $path = preg_replace(
    '/(\p{Lu})/'," $1",$path);
  $path = strtr($path,array(' ' =>'-'));
  $path = strtolower($path);
  $path = ltrim($path,'-');
  $path = strtr($path,array(
    '.-' => '.','/-' => '/'));
  if($EnablePathInfo == 1)
    $path = str_replace('.','/',$path); 
  $path = strtr($path,$RouterCamelToDashedReserved);    
  return $path;
}

/// Redirects to path or $line['Out']
function RouterFnRedirect($path,$line) {
  if('In'!=$line['Direction']){
    return $path;
  }
  if($path == @$line['In']){
    header('Location: '.@$line['Out'],TRUE,301);
  }
  return $path;
}

